@@ -1,12 +1,44 @@
 /* vim: set sw=4 ts=4 sts=4 et: */
+
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+#include <stdio.h>
+#include <string.h>
+#include <ctype.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <sys/mman.h>
+
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
+#elif defined __GNUC__
+# define alloca __builtin_alloca
+#elif defined _AIX
+# define alloca __alloca
+#elif defined _MSC_VER
+# include <malloc.h>
+# define alloca _alloca
+#else
+# include <stddef.h>
+# ifdef  __cplusplus
+extern "C"
+# endif
+void *alloca (size_t);
+#endif
+
 #include "Efreet.h"
 #include "efreet_private.h"
 
-static Ecore_Hash *efreet_ini_parse(const char *file);
+static Eina_Hash *efreet_ini_parse(const char *file);
 static char *efreet_ini_unescape(const char *str);
 
-static void efreet_ini_section_save(Ecore_Hash_Node *node, FILE *f);
-static void efreet_ini_value_save(Ecore_Hash_Node *node, FILE *f);
+static Eina_Bool
+efreet_ini_section_save(const Eina_Hash *hash, const void *key, void *data, void *fdata);
+static Eina_Bool
+efreet_ini_value_save(const Eina_Hash *hash, const void *key, void *data, void *fdata);
 
 /**
  * The number of times the Ini subsytem has been initialized
@@ -22,7 +54,7 @@ int
 efreet_ini_init(void)
 {
     if (init++) return init;
-    if (!ecore_string_init()) return --init;
+    if (!eina_stringshare_init()) return --init;
     return init;
 }
 
@@ -35,7 +67,7 @@ int
 efreet_ini_shutdown(void)
 {
     if (--init) return init;
-    ecore_string_shutdown();
+    eina_stringshare_shutdown();
     return init;
 }
 
@@ -65,23 +97,28 @@ efreet_ini_new(const char *file)
 /**
  * @internal
  * @param file The file to parse
- * @return Returns an Ecore_Hash with the contents of @a file, or NULL if the
+ * @return Returns an Eina_Hash with the contents of @a file, or NULL if the
  *         file fails to parse or if the file doesn't exist
- * @brief Parses the ini file @a file into an Ecore_Hash
+ * @brief Parses the ini file @a file into an Eina_Hash
  */
-static Ecore_Hash *
+static Eina_Hash *
 efreet_ini_parse(const char *file)
 {
     const char *buffer, *line_start;
     FILE *f;
-    Ecore_Hash *data, *section = NULL;
+    Eina_Hash *data, *section = NULL;
     struct stat file_stat;
     int line_length, left;
 
     f = fopen(file, "rb");
     if (!f) return NULL;
 
-    if (fstat(fileno(f), &file_stat) || file_stat.st_size < 1)
+    if (fstat(fileno(f), &file_stat) || (file_stat.st_size < 1))
+    {
+        fclose(f);
+        return NULL;
+    }
+    if (!S_ISREG(file_stat.st_mode)) /* if not a regular file - close */
     {
         fclose(f);
         return NULL;
@@ -89,15 +126,13 @@ efreet_ini_parse(const char *file)
 
     left = file_stat.st_size;
     buffer = mmap(NULL, left, PROT_READ, MAP_SHARED, fileno(f), 0);
-    if (!buffer)
+    if (buffer == MAP_FAILED)
     {
         fclose(f);
         return NULL;
     }
 
-    data = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-    ecore_hash_free_key_cb_set(data, ECORE_FREE_CB(ecore_string_release));
-    ecore_hash_free_value_cb_set(data, ECORE_FREE_CB(ecore_hash_destroy));
+    data = eina_hash_string_small_new(EINA_FREE_CB(eina_hash_free));
 
     line_start = buffer;
     while (left > 0)
@@ -107,7 +142,7 @@ efreet_ini_parse(const char *file)
         /* find the end of line */
         for (line_length = 0; 
                 (line_length < left) && 
-                (line_start[line_length] != '\n'); ++line_length)
+                (line_start[line_length] != '\n'); line_length++)
             ;
 
         /* check for all white space */
@@ -136,7 +171,6 @@ efreet_ini_parse(const char *file)
 
             if (line_start[header_length] == ']')
             {
-                Ecore_Hash *old;
                 const char *header;
 
                 header = alloca(header_length * sizeof(unsigned char));
@@ -145,31 +179,26 @@ efreet_ini_parse(const char *file)
                 memcpy((char*)header, line_start + 1, header_length - 1);
                 ((char*)header)[header_length - 1] = '\0';
 
-                section = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-                ecore_hash_free_key_cb_set(section,
-                            ECORE_FREE_CB(ecore_string_release));
-                ecore_hash_free_value_cb_set(section, ECORE_FREE_CB(free));
+                section = eina_hash_string_small_new(free);
 
-                old = ecore_hash_remove(data, header);
-                if (old) printf("[efreet] Warning: duplicate section '%s' "
-                                "in file '%s'\n", header, file);
+                eina_hash_del(data, header, NULL);
+//                if (old) printf("[efreet] Warning: duplicate section '%s' "
+  //                              "in file '%s'\n", header, file);
 
-                IF_FREE_HASH(old);
-                ecore_hash_set(data, (void *)ecore_string_instance(header),
-                                section);
+                eina_hash_add(data, header, section);
             }
             else
             {
                 /* invalid file - skip line? or refuse to parse file? */
                 /* just printf for now till we figure out what to do */
-                printf("Invalid file (%s) (missing ] on group name)\n", file);
+//                printf("Invalid file (%s) (missing ] on group name)\n", file);
             }
             goto next_line;
         }
 
         if (section == NULL)
         {
-            printf("Invalid file (%s) (missing section)\n", file);
+//            printf("Invalid file (%s) (missing section)\n", file);
             goto next_line;
         }
 
@@ -180,7 +209,6 @@ efreet_ini_parse(const char *file)
         if (sep < line_length)
         {
             const char *key, *value;
-            char *old;
             int key_end, value_start, value_end;
 
             /* trim whitespace from end of key */
@@ -191,16 +219,16 @@ efreet_ini_parse(const char *file)
             if (!isspace(line_start[key_end])) key_end++;
 
             /* trim whitespace from start of value */
-            for (value_start = sep + 1; 
-                    (value_start < line_length) && 
-                    isspace(line_start[value_start]); ++value_start)
+            for (value_start = sep + 1;
+                 (value_start < line_length) &&
+                 isspace(line_start[value_start]); ++value_start)
                 ;
 
             /* trim \n off of end of value */
             for (value_end = line_length; 
-                    (value_end > value_start) &&
-                    ((line_start[value_end] == '\n') ||
-                        (line_start[value_end] == '\r')); --value_end)
+                 (value_end > value_start) &&
+                 ((line_start[value_end] == '\n') ||
+                  (line_start[value_end] == '\r')); --value_end)
                 ;
 
             if (line_start[value_end] != '\n'
@@ -212,7 +240,7 @@ efreet_ini_parse(const char *file)
             if (key_end == 0)
             {
                 /* invalid file... */
-                printf("Invalid file (%s) (invalid key=value pair)\n", file);
+//                printf("Invalid file (%s) (invalid key=value pair)\n", file);
 
                 goto next_line;
             }
@@ -228,23 +256,19 @@ efreet_ini_parse(const char *file)
                     value_end - value_start);
             ((char*)value)[value_end - value_start] = '\0';
 
-            old = ecore_hash_remove(section, key);
-            IF_FREE(old);
-
-            ecore_hash_set(section, (void *)ecore_string_instance(key),
-                           efreet_ini_unescape(value));
-        }
-        else
-        {
-            /* invalid file... */
-            printf("Invalid file (%s) (missing = from key=value pair)\n", file);
+            eina_hash_del(section, key, NULL);
+            eina_hash_add(section, key, efreet_ini_unescape(value));
         }
+//        else
+//        {
+//            /* invalid file... */
+//            printf("Invalid file (%s) (missing = from key=value pair)\n", file);
+//        }
 
 next_line:
         left -= line_length + 1;
         line_start += line_length + 1;
     }
-
     munmap((char*) buffer, file_stat.st_size);
     fclose(f);
 
@@ -279,7 +303,7 @@ efreet_ini_save(Efreet_Ini *ini, const char *file)
 
     f = fopen(file, "wb");
     if (!f) return 0;
-    ecore_hash_for_each_node(ini->data, ECORE_FOR_EACH(efreet_ini_section_save), f);
+    eina_hash_foreach(ini->data, efreet_ini_section_save, f);
     fclose(f);
 
     return 1;
@@ -296,7 +320,7 @@ efreet_ini_section_set(Efreet_Ini *ini, const char *section)
 {
     if (!ini || !ini->data || !section) return 0;
 
-    ini->section = ecore_hash_get(ini->data, section);
+    ini->section = eina_hash_find(ini->data, section);
     return (ini->section ? 1 : 0);
 }
 
@@ -309,22 +333,16 @@ efreet_ini_section_set(Efreet_Ini *ini, const char *section)
 EAPI void
 efreet_ini_section_add(Efreet_Ini *ini, const char *section)
 {
-    Ecore_Hash *hash;
+    Eina_Hash *hash;
 
     if (!ini || !section) return;
 
     if (!ini->data)
-    {
-        ini->data = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-        ecore_hash_free_key_cb_set(ini->data, ECORE_FREE_CB(ecore_string_release));
-        ecore_hash_free_value_cb_set(ini->data, ECORE_FREE_CB(ecore_hash_destroy));
-    }
-    if (ecore_hash_get(ini->data, section)) return;
+      ini->data = eina_hash_string_small_new(EINA_FREE_CB(eina_hash_free));
+    if (eina_hash_find(ini->data, section)) return;
 
-    hash = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-    ecore_hash_free_key_cb_set(hash, ECORE_FREE_CB(ecore_string_release));
-    ecore_hash_free_value_cb_set(hash, ECORE_FREE_CB(free));
-    ecore_hash_set(ini->data, (void *)ecore_string_instance(section), hash);
+    hash = eina_hash_string_small_new(free);
+    eina_hash_add(ini->data, section, hash);
 }
 
 /**
@@ -339,7 +357,7 @@ efreet_ini_string_get(Efreet_Ini *ini, const char *key)
 {
     if (!ini || !key || !ini->section) return NULL;
 
-    return ecore_hash_get(ini->section, key);
+    return eina_hash_find(ini->section, key);
 }
 
 /**
@@ -354,7 +372,8 @@ efreet_ini_string_set(Efreet_Ini *ini, const char *key, const char *value)
 {
     if (!ini || !key || !ini->section) return;
 
-    ecore_hash_set(ini->section, (void *)ecore_string_instance(key), strdup(value));
+    eina_hash_del(ini->section, key, NULL);
+    eina_hash_add(ini->section, key, strdup(value));
 }
 
 /**
@@ -505,29 +524,28 @@ efreet_ini_localestring_get(Efreet_Ini *ini, const char *key)
     {
         snprintf(buf, maxlen, "%s[%s_%s@%s]", key, lang, country, modifier);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang && country)
     {
         snprintf(buf, maxlen, "%s[%s_%s]", key, lang, country);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang && modifier)
     {
         snprintf(buf, maxlen, "%s[%s@%s]", key, lang, modifier);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang)
     {
         snprintf(buf, maxlen, "%s[%s]", key, lang);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
-
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found)
@@ -634,15 +652,21 @@ efreet_ini_unescape(const char *str)
     return buf;
 }
 
-static void
-efreet_ini_section_save(Ecore_Hash_Node *node, FILE *f)
+static Eina_Bool
+efreet_ini_section_save(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    fprintf(f, "[%s]\n", (char *)node->key);
-    ecore_hash_for_each_node(node->value, ECORE_FOR_EACH(efreet_ini_value_save), f);
+    FILE *f = fdata;
+
+    fprintf(f, "[%s]\n", (char *)key);
+    eina_hash_foreach(value, efreet_ini_value_save, f);
+    return EINA_TRUE;
 }
 
-static void
-efreet_ini_value_save(Ecore_Hash_Node *node, FILE *f)
+static Eina_Bool
+efreet_ini_value_save(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    fprintf(f, "%s=%s\n", (char *)node->key, (char *)node->value);
+    FILE *f = fdata;
+
+    fprintf(f, "%s=%s\n", (char *)key, (char *)value);
+    return EINA_TRUE;
 }
